# $Id: search.gap,v 1.22 2001/07/16 03:45:53 nathand Exp $
#
# IMPORTANT NOTE:  The basic subgroup searching functions have been
# replaced with better versions in rigour.gap
# 
# p-group searcher:

Read("reps.gap");

FindPGroupCover := function(M)
  local factors, c, p, lastSize,f, A, i,rank, l, G ;
  factors := [];
  for c in M.Homology do
    for p in FactorsInt(c) do 
      AddSet(factors, p);
    od;
  od;
 
  G := M.Group();
  for p in factors do
    lastSize := 0;
    for i in [1..100] do
       f := EpimorphismPGroup(G, p, i);
       A := Image(f);
       if Size(A) > 1200 or Size(A) = lastSize then
          break;
       fi;
       lastSize := Size(A);
       Print(M.Name," ", "p = ", p, " i = ", i, " rank : ");
       rank := AbelianRankOfSubgroup(G, Kernel(f));
       Print(rank, "\n");
       if rank > 0 then
         FoundPosBettiCover(M, [ p, i], "p-group");
         break;
       fi;
   od;
   
 od;
end;

FindPGroupCover2 := function(G)
  local factors, c, p, lastSize,f, A, i,rank, l;
  factors := [];
  for c in AbelianInvariants(G) do
    for p in FactorsInt(c) do 
      AddSet(factors, p);
    od;
  od;
 
  for p in factors do
    lastSize := 0;
    for i in [1..100] do
       f := EpimorphismPGroup(G, p, i);
       A := Image(f);
       if Size(A) > 1200 or Size(A) = lastSize then
          break;
       fi;
       lastSize := Size(A);
       Print("p = ", p, " i = ", i, " rank : ");
       rank := AbelianRankOfSubgroup(G, Kernel(f));
       Print(rank, "\n");
   od;
 od;
end;

SmallestNecessarySubgroup := function(G)
  local C, U, XX, T;
  T := CharacterTable(G);
  C := List(ConjugacyClassesSubgroups(G), h -> Representative(h));
  Sort(C, function(g, h) return  Order(h) <= Order(g); end);
  for U in C do
     XX := PermutationCharacter(G, U);
     if not 0 in List( Irr(T), y-> ScalarProduct(XX,y) ) then
       return U;
     fi;
  od;
end;

PSL2PCoverFinder := function(M, p, Q, U)
  local G, f, rank;
  #Print(M.Name, "\n");
  G := M.Group();
  for f in GQuotients(G, Q) do
       Print(M.Name," has PSL(2," , p, ") cover w rank: ");
       rank := AbelianRankOfSubgroup(G, PreImage(f, U));
       Print(rank, "\n");
       if rank > 0 then
         FoundPosBettiCover(M, ["PSL", p], "PSL2PCoverFinder");
         break;
       fi;
  od;
end;


PSLCoverTester := function(M, p, Q, U)
  local G, f, rank;
  #Print(M.Name, "\n");
  G := M.Group();
  for f in GQuotients(G, Q) do
       Print(M.Name," has PSL(2," , p, ") cover w rank: ");
       rank := AbelianRankOfSubgroup(G, PreImage(f, U));
       Print(rank, "\n");
       if rank > 0 then
         #FoundPosBettiCover(M, ["PSL", p], "PSL2PCoverFinder");
         break;
       fi;
  od;
end;


PSLCoverTester2 := function(M, p, Q, U, V, W)
  local G, f, rank1, rank2, rank3;
  #Print(M.Name, "\n");
  G := M.Group();
  for f in GQuotients(G, Q) do
       Print(M.Name," has PSL(2," , p, ") cover w rank: ");
       rank1 := AbelianRankOfSubgroup(G, PreImage(f, U));
       rank2 := AbelianRankOfSubgroup(G, PreImage(f, V));
       rank3 := AbelianRankOfSubgroup(G, PreImage(f, W));
       Print(rank1, " ", rank2, " ", rank1 - rank2," ", rank3, "\n");
       #if rank > 0 then
         #FoundPosBettiCover(M, ["PSL", p], "PSL2PCoverFinder");
       #  break;
       #fi;
  od;
end;

#  Function for testing if we could compute the rank of the matrix
#  rigerously.

EstimateBits := function(M)
  local p, G, Q, U, f, A;
  if M.KnownPosBettiCover and Length(M.GoodCover) = 2 then 
     G := M.Group();
     p := M.GoodCover[2];
     Q := PSL(2, p);
     U := SmallestNecessarySubgroup(Q);
     f := GQuotients(G, Q)[1];
     A:=RelatorMatrixAbelianizedSubgroupRrs(G, PreImage(f, U));
     Print(M.Name, " p = ", p, " bound = ", Hadamard(A), "\n");
  fi;
end;

# takes the given subgroup, and then examines the homology of that
# subgroup via representations of Sn, where n = Index(G, H).

CheckDegreeNCover := function(G, H, RepsOfSn)
  local h, fake_rep, rep, rank, count;
  count := 0;
  h := FactorCosetAction(G, H);
  for fake_rep in RepsOfSn do
    rep := fake_rep(SymmetricGroup(Index(G,H)));
    rank := QuickRankOfHomologyWithLocalCoeffs(G, 
        List(GeneratorsOfGroup(G), g -> Image(rep, Image(h, g))));
    Print(rank, " ");	
    if (count mod 5) = 0 then Print("\n"); fi;
    if rank > 0 then
       break;
    fi;
    count := count + 1;
  od;
  Print("\n"); 
  return [rank > 0, Position(RepsOfSn, rep)];
end;

TypeOfGroup := function(G)
   if IsSolvable(G) then return "solvable"; fi;
   if IsSimple(G) then return "simple"; fi;
   return "mixed";
end;

CheckDegreeNineCovers := function(M, RepsOfS9)
   local G, H, a,f,imagef, sgs ;
   G := M.Group();
   sgs := LowIndexSubgroupsFpGroup(G, TrivialSubgroup(G), 10);
   for H in  Filtered(sgs, h->Index(G,h) = 9) do 
      Print("   Examining Cover of index 9\n");
      a := CheckDegreeNCover(G, H, RepsOfS9);
      if a[1] then 
         f :=  FactorCosetAction(G, H);
	 imagef := Image(f);
         FoundPosBettiCover(M, ["NLowIndex(9)", Size(imagef),
	  TypeOfGroup(imagef), a[2]], "NMD, quickrank");
 	 return;
      fi;
   od;

end;
               
CheckIntersectionsOfDegreeNCovers := function(M, n, RepsOfSn)
  local G, sgs, H1, H2, n1, n2, r1, r2, count, 
    rank, ImagesOfGen, h1, h2, d1, d2, rind1, rind2;
  G := M.Group();
  sgs := Filtered( LowIndexSubgroupsFpGroup(G, TrivialSubgroup(G), n),
    h->Index(G,h) > 2);
  Print(sgs, "\n");
  for H1 in sgs do
    for H2 in sgs do
      if H2 <> H1 then
        n1 := Index(G, H1);
        n2 := Index(G, H2);
        Print("Examining subgroups of index ", [n1, n2], "\n     ");
        count := 0;
        for rind1 in [1..Length(RepsOfSn[n1])] do
           r1 := RepsOfSn[n1][rind1](SymmetricGroup(n1));
           for rind2 in [1..Length(RepsOfSn[n2])] do
              r2 := RepsOfSn[n2][rind2](SymmetricGroup(n2));
              h1 := FactorCosetAction(G, H1);
              h2 := FactorCosetAction(G, H2);
              d1 := DimensionsMat(Image(r1, (1,2)))[1];
              d2 := DimensionsMat(Image(r2, (1,2)))[1];
              ImagesOfGen := List(GeneratorsOfGroup(G), 
                g-> TensorOfMatrices(Image(r1, Image(h1, g)), 
                 Image(r2, Image(h2, g))) );
	      Print("\n");
              rank := QuickRankOfHomologyWithLocalCoeffs(G, ImagesOfGen);
              Print([[rind1, d1],
                     [rind2, d2], rank], " ");	
              if (count mod 5) = 0 then Print("\n     "); fi;
              if rank > 0 then
                FoundPosBettiCover(M, ["IntersectionOfSubgroups", [n1, n2],
	         [d1, d2], rank], "");
                 #return;
              fi;
              count := count + 1;
           od;
         od;
       fi;
     od;
   od;
end;

#
# The function 

CheckTensorsOfRepsOfCosetActionOfLowIndex := function(M, MaxSubgroupIndex,
    MaxDimensionOfRepresentation, MaxNumberOfSubgroupSubsets, MaxSizeOfSubset, 
    MaxNumberOfRepresentations, RepsOfSn, PartOfRepsOfSn, 
    PrintEveryRepConsidered, StopWhenPosBettiFound)

# does the following: Let G = pi_1(M).  The function looks at
# subgroups H_1,...,H_k of G index at most MaxSubgroupIndex.  For each
# H_i, it looks at the induced homomorphims to S_n and composes that
# homomorphism with the various irreducible representations of S_n to
# yield a large collection of representations of G.  It then forms the
# tensor products of those representatations rho_1 x ... x rho_k where
# the rho_i come from distinct H_i, and the total dimension of rho_1 x
# ... x rho_k is less than MaxDimensionOfRepresentation.  It then
# computes the first betti number of G w.r.t to these representations,
# sorted by dimension from smallest to largest.
#
# The arguements MaxNumberOfSubgroupSubsets, MaxSizeOfSubset and MaxNumberOfRepresentations
# are there to limit the amout of time the function spends on any given
# manifold.  If there are, say, 20 subgroups then clearly we can't look at
# all the tensor products as there could easily be millions. 

   local G, H, IndexH, Reps, TempReps, i, HReps, j, rho, dim_rho, TensoredReps, TempTensoredReps,
   TensoredRep, rank,h, g, ImagesOfGen, s, comb;
  
   G := M.Group();
   H := Filtered(LowIndexSubgroupsFpGroup(G, TrivialSubgroup(G), MaxSubgroupIndex),
                         h -> Index(G,h) <> 1);
   if Length(H) = 0 then Print("no subgroups\n"); return; fi;
   IndexH := List(H, h-> Index(G, h));
   Reps := [];
   for i in [1..Length(H)] do
     HReps := [];
     h := FactorCosetAction(G, H[i]);
     for j in [1..Length(RepsOfSn[IndexH[i]])] do
       rho := RepsOfSn[IndexH[i]][j];
       dim_rho :=  DimensionsMat(Image(rho, (1,2)))[1];
       if dim_rho <= MaxDimensionOfRepresentation then
          Add(HReps, [dim_rho, [i,IndexH[i], PartOfRepsOfSn[IndexH[i]][j], dim_rho],  
             List(GeneratorsOfGroup(G), g -> Image(rho, Image(h, g))) ] );
       fi;
     od;
     Add(Reps, HReps);
   od;

   # We have to be careful how many sets of subgroups we look at.
   if Length(H) <= MaxSizeOfSubset then
       s := Length(H);
   else
     for s in [1..MaxSizeOfSubset] do
       if NrCombinations([1..Length(H)], s) > MaxNumberOfSubgroupSubsets and s > 1 then
         s := s - 1;
         break;
      fi;
    od;
   fi;
   
   if s < Length(H) and s < MaxSizeOfSubset then Print("Not all subsets of subgroups considered -- too many.\n"); fi;

   TensoredReps := [];
   for comb in Combinations([1..Length(H)], s) do
      TempReps := List(comb, ind -> Reps[ind]);
      TempTensoredReps := List(Cartesian(TempReps),  
             rho -> [Product(List(rho, rho_i -> rho_i[1])), rho]);
      TempTensoredReps := Filtered(TempTensoredReps, 
             rho -> rho[1] <= MaxDimensionOfRepresentation);
      Append(TensoredReps, TempTensoredReps);
   od;
   Sort(TensoredReps, function(x,y) return x[1] < y[1]; end);
   
   #Make sure that we don't look at too many representations
   if Length(TensoredReps) > MaxNumberOfRepresentations then
      Print("Too many reps: ", [Length(TensoredReps) ,
            MaxNumberOfRepresentations], " some dropped\n");
      TensoredReps := TensoredReps{[1..MaxNumberOfRepresentations]};
   fi;

   for TensoredRep in TensoredReps do
     if PrintEveryRepConsidered then
       Print("Exam ", List(TensoredRep[2], x->x[2]), "; Dim ",
          TensoredRep[1], "\n");
     fi;
     ImagesOfGen := List([1..Length(GeneratorsOfGroup(G))], i -> 
         TensorListOfMatrices(List(TensoredRep[2], y -> y[3][i] )));
     rank := QuickRankOfHomologyWithLocalCoeffs(G, ImagesOfGen);
	
     if rank > 0 then
       if not PrintEveryRepConsidered then
       Print("Exam ", List(TensoredRep[2], x->x[2]), "; Dim ",
          TensoredRep[1], "\n");
       fi;  
       Print("Rank: ", rank, "\n");
       # If we're in !StopWhenPosBettiFound we want to be careful 
       # not to overwrite existing info, since we prob. want the smallest dim rep.
       if not M.KnownPosBettiCover then
          FoundPosBettiCover(M,
             ["TensorOfReps", List(TensoredRep[2], x-> x[2]), rank], "TensorOfReps");
       fi;
       if StopWhenPosBettiFound then
         return;
       fi;
     fi;
   od;
end;
   
#---------------------------------------------------------------------------
#
#    FINDING COVERS FOR USE IN STATISTICS:
#
#---------------------------------------------------------------------------   

FindQuotientsOfManifolds := function(manifolds, Q, type, ident)
    local M, G, f;
    Print(Q, "\n");
    for M in manifolds do
       Print(M.Name, "\n");
       G := M.Group();
       for f in GQuotients(G, Q) do
         Print(f, "\n");
         Add(M.Quotients,  
          rec(Type := type, Ident := ident, Name := Name(Q),
              Rank := "unknown",
              Image := StoreGroupHom(G, f)  
             ) 
          );
       od; # for f in GQuotients
    od; # for M in manifolds
end;


# The following function tests for, and stores a homomorphism to a 
# simple group
         
TestSimpleGroup := function(manifolds, ident)
    local Q;
    Q := SimpleGroup(ident);
    FindQuotientsOfManifolds(manifolds, Q, "simple", ident);
end;

TestPerfectGroup := function(manifolds, ident)
    local Q;
    Q := PerfectGroup(IsPermGroup, ident);
    FindQuotientsOfManifolds(manifolds, Q, "perfect", ident);
end;

ComputeRankOfCover := function(manifold, quo_rec, Q, U)
   local G, f;
   # check to see if already computed, in which case we skip
   if IsInt(quo_rec.Rank) then return; fi;
   # check to see if we crashed earlier
   if quo_rec.Rank = fail then return; fi;
   # in case we crash
   quo_rec.Rank := fail;
   G := manifold.Group();
   f := GroupHomomorphismByImages(G, Q, 
         GeneratorsOfGroup(G), GeneratorsOfGroup(quo_rec.Image));
   Print(manifold.Name, "  ", quo_rec.Name, "\n");
   quo_rec.Rank :=  AbelianRankOfSubgroup(G, PreImage(f, U));
end;

ComputeRanksOfSimpleCovers := function(manifolds, ident, Q, U)
   local M, quo_rec;
   for M in manifolds do
     for quo_rec in M.Quotients do
	 if quo_rec.Type = "simple" and quo_rec.Ident = ident then
             ComputeRankOfCover(M, quo_rec, Q, U);
          fi;
     od;
   od;
end;
   
EasyComputeRanksOfSimpleCovers := function(manifolds)
    local ident, Q, U, M, need_to_do, quo_rec;
    for ident in IdentifiersOfSimpleGroups() do
       need_to_do := false;
       for M in manifolds do
         for quo_rec in M.Quotients do
	    if quo_rec.Type = "simple" and quo_rec.Ident = ident 
               and  quo_rec.Rank = "unknown" then
                need_to_do := true;
                break;
            fi;
         od;
         if need_to_do then break; fi;
       od;
       Print(ident, "  ", need_to_do, "\n");
       if need_to_do then 
          Q := SimpleGroup(ident);
          U := SmallestNecessarySubgroup(Q);
          ComputeRanksOfSimpleCovers(manifolds, ident, Q, U);
       fi;
    od;
end;

EasyComputeRankOfCovers := function(manifold)
   local G, Q, U, quo_rec;
   G := manifold.Group();
   for quo_rec in manifold.Quotients do
     if quo_rec.Rank = "unknown" then 
        Q := quo_rec.Image;
        U := SmallestNecessarySubgroup(Q);
        ComputeRankOfCover(manifold, quo_rec, Q, U);
     fi;
   od;
end;
    

UpdateGoodCoverBasedOnSimpleCovers := function(manifold)
  local Q;
  if not manifold.KnownPosBettiCover then 
     for Q in manifold.Quotients do 
       if IsInt(Q.Rank) and Q.Rank > 0 then
          FoundPosBettiCover(manifold, ["Simple", Q.Ident], "quickrank");
	  break;
       fi;
     od;
   fi;
 end;
                   
# Let Q be the SimpleGroup(ident).  The function returns
# [list of ranks of betti numbers of Q-covers]

GetSimpleGroupQuotientData := function(manifold, ident)
   local quo_rec, quo_data;
   quo_data := [];
   for quo_rec in manifold.Quotients do
     if quo_rec.Type = "simple" and quo_rec.Ident = ident then
        if quo_rec.Rank <> fail and quo_rec.Rank <> "unknown" then 
            Add(quo_data, quo_rec.Rank);
        fi;
     fi;
   od;
   return quo_data;
end;

EasySimpleGroupQuotientStat := function(manifold, ident)
   local L;
   L := GetSimpleGroupQuotientData(manifold, ident);
   if Length(L) > 0 then
      if Maximum(L) > 0 then 
         return 1;
      else 
         return -1;
      fi;
   fi;
   return 0;
end;

GetSimpleQuotientKernel := function(manifold, quo_rec)
   local G, Q, U, f;   
   G := manifold.Group();
   Q := quo_rec.Image;
   U := SmallestNecessarySubgroup(Q);
   f := GroupHomomorphismByImages(G, Q, 
         GeneratorsOfGroup(G), GeneratorsOfGroup(quo_rec.Image));
   return [G, PreImage(f, U)];
end;

GetSimpleQuotient := function(manifold, quo_rec)
   local G, Q, U, f;   
   G := manifold.Group();
   Q := quo_rec.Image;
   f := GroupHomomorphismByImages(G, Q, 
         GeneratorsOfGroup(G), GeneratorsOfGroup(quo_rec.Image));
   return [G, Q, f];
end;

GetSimpleQuotientRecordsOfManifold := function(manifold, ident)
  local quotients, quo_rec;
  quotients := [];
  for quo_rec in manifold.Quotients do
	 if quo_rec.Type = "simple" and quo_rec.Ident = ident then
	   Add(quotients, quo_rec);
         fi;
  od;
  return quotients;
end;    

GetSimpleQuotientRecordsOfManifolds := function(manifolds, ident)
   local  quotients, M, d;
   quotients := [];
   for M in manifolds do
     for d in  GetSimpleQuotientRecordsOfManifold(M, ident) do
       Add(quotients, [M, d]);
     od;
   od;
   return quotients;
end;

GetSimpleQuotientRanksOfManifolds := function(manifolds, ident)
   local ranks, M;
   ranks := [];
   for M in manifolds do
       Append(ranks, GetSimpleGroupQuotientData(M, ident));
   od;
   Sort(ranks);
   return ranks;
end;

# the following function takes a list of manifolds and produces
# a list of the form:
# 
# [ simple_gp_ident,  [test_results] ]
# 
# for all SimpleGroups in order until 3 in a row are found for which the data 
# is all 0.


DataOnSimpleQuotients := function(manifolds)
  local data, group_data, ident, missing, hits, covers;
  data := [];
  missing := 0;
  for ident in IdentifiersOfSimpleGroups() do 
     group_data := List(manifolds, M -> EasySimpleGroupQuotientStat(M, ident));
     if Maximum(group_data) = 0 and Minimum(group_data) = 0 then
        missing := missing + 1;
	if missing > 3 then 
           break;
        fi;
     else
        missing := 0;     
        hits := Number(group_data, x -> x = 1);
        covers := Number(group_data, x -> x = 1 or x = -1);
        Add(data, [ident,
           [hits/Length(group_data), covers/Length(group_data), hits/covers],
           group_data]);
      fi;
  od;
  return data;
end;

# for saving to python

SaveDataOnSimpleQuotients := function(filename, manifolds)
  local data, group_data, ident, missing, hits, covers;
  data := [];
  missing := 0;
  for ident in IdentifiersOfSimpleGroups() do 
     group_data := List(manifolds, M -> EasySimpleGroupQuotientStat(M, ident));
     if Maximum(group_data) = 0 and Minimum(group_data) = 0 then
        missing := missing + 1;
	if missing > 3 then 
           break;
        fi;
     else
        missing := 0;     
        Add(data, [ident,
           SimpleGroupName(ident),
           group_data]);
      fi;
  od;
  PrintTo(filename, data);
end;

# The thing below is very, very, very stupid.

PrimativeCorrelation := function(list1, list2)
  local data;
  Assert(0, Length(list1) = Length(list2));
  data := List([1..Length(list1)], i -> list1[i]*list2[i]);
  return Sum(data)/Number(data, d -> d <> 0);
end;

Correlations := function(data)
  local n, real_data, cor;
  n := Length(data);
  real_data := List(data, x -> x[3]);
  return  List(real_data, dat1 -> 
     List(real_data, dat2 -> PrimativeCorrelation(dat1,dat2))); 
end;


# for a number in [-1,1] 

FixedLengthThreeDigitDecimal := function(frac)
   local s, t;
   if frac < 0 then s := "-"; else s := " "; fi;
   t := String(AbsInt(Int(frac*1000)));
   Assert(0, Length(t) < 5);
   while Length(t) < 4 do
     t := Concatenation("0",t);
   od;
   return Concatenation(s, t{[1]}, ".", t{[2..4]});
end;
  
Percentage := function(frac)
   local t;
   Assert(0, 0 <= frac and frac <= 100);
   t := String(Int(frac*1000));
   while Length(t) < 3 do
     t := Concatenation(" ",t);
   od;
   return Concatenation(t{[1..2]}, ".", t{[3]});
end;   

FormatNumber := function(frac)
   local t;
   Assert(0, 0 <= frac and frac <= 100);
   t := String(Int(frac*1000));
   while Length(t) < 3 do
     t := Concatenation(" ",t);
   od;
   if frac >= 10 then
   return Concatenation(t{[1..2]}, ".", t{[3]});
   fi;
   if frac  < 10 then
   return Concatenation(t{[1]}, ".", t{[2..3]});
   fi;
    
end;   


DisplayCorrelations := function(data)
   local cor, c, i, name;
   cor := Correlations(data);
   #print data heads across
   Print("        ");
   Print("Hit  HavCov SucRat     ");
   for i in [1..Length(data)] do 
     name := SimpleGroupName(data[i][1]);
     while Length(name) < 6 do 
       name := Concatenation(name, " ");
     od;
     Print(name, "  ");
   od;
   Print("\n");

   for i in [1..Length(data)] do 
     name := SimpleGroupName(data[i][1]);
     while Length(name) < 6 do 
       name := Concatenation(name, " ");
     od;
     Print(name, "  ");
     Print(Percentage(data[i][2][1]), "  ", Percentage(data[i][2][2]), "   ",
         Percentage(data[i][2][3]), "     ");
     for c in cor[i] do
       Print(FixedLengthThreeDigitDecimal(c), "  ");
     od;
     Print("\n");
   od;
end;

# just look at if having a cover of type blah tells us anything
# about having a cover of type bleeh.

DisplayCoverCorrelations := function(data)
  local f, CD;
  f := function(x) if x = 0 then return -1; else return 1; fi; end;
  CD := List(data, x -> [x[1], x[2], List(x[3], f)]);
  DisplayCorrelations(CD);
end;
   

DisplayHitCorrelations := function(data)
  local f, CD;
  f := function(x) if x = 1 then return 1; else return -1; fi; end;
  CD := List(data, x -> [x[1], x[2], List(x[3], f)]);
  DisplayCorrelations(CD);
end;
   
SmallestSimpleCover := function(manifold)
   local ident, quo_rec;
   for ident in IdentifiersOfSimpleGroups() do
      for quo_rec in manifold.Quotients do
	 if quo_rec.Type = "simple" and quo_rec.Ident = ident then
             if quo_rec.Rank > 0 then 
                 return ident;
             fi;
          fi;
     od;
   od;
   return fail;
end;

SmallestPSL2Cover := function(manifold)
   local ident, quo_rec;
   for ident in IdentifiersOfNonL2Groups() do
      for quo_rec in manifold.Quotients do
	 if quo_rec.Type = "simple" and quo_rec.Ident = ident then
             if quo_rec.Rank > 0 then 
                 return ident;
             fi;
          fi;
     od;
   od;
   return fail;
end;


SmallestNonPSL2Cover := function(manifold)
   local ident, quo_rec, PSLsizes;
   PSLsizes := IdentifiersOfNonL2Groups();
   for ident in IdentifiersOfSimpleGroups() do
      for quo_rec in manifold.Quotients do
	 if quo_rec.Type = "simple" and quo_rec.Ident = ident then
             if quo_rec.Rank > 0 and not (ident in PSLsizes) then 
                 return ident[1];
             fi;
          fi;
     od;
   od;
   return fail;
end;

#L := List(CM,  SmallestSimpleCover);
#Id := IdentifiersOfSimpleGroups();; Add(Id, fail);       
#List(Id, ident -> Number(L, l -> l = ident));

#----------------------------------------------------------------------
# 
# Below functions for computing modular representations of a given
# finite group
#
#-----------------------------------------------------------------------

ModularRepresentations := function(G, p)
   local H, f, M, Irred;
   H := SmallestNecessarySubgroup(G);
   f := FactorCosetAction(G,H);
   M := GModuleByMats(List(GeneratorsOfGroup(G),
           g ->  PermutationMat(Image(f, g), Index(G,H))*Z(p)^0), GF(p));
   Irred := MTX.CollectedFactors(M);
   return List(Irred, I -> MTX.Generators(I[1]));
end;

ComputeTwistedHomology := function(manifold, Q, quo_rec, representation)
  local ImagesOfGens, G, f, h, M;
  G := manifold.Group();
  f := GroupHomomorphismByImages(G, Q, GeneratorsOfGroup(G), GeneratorsOfGroup(quo_rec.Image));
  M := Group(representation);
  h := GroupHomomorphismByImages(Q, M, GeneratorsOfGroup(Q), representation);
  ImagesOfGens := List(GeneratorsOfGroup(G), g -> Image(h, Image(f, g)));
  return RankOfHomologyWithLocalCoeffs(G, ImagesOfGens);
end;

ComputeTwistedHomologyAlt := function(manifold, Q, quo_rec, rep)
  local ImagesOfGens, G, f, h, M;
  G := manifold.Group();
  f := GroupHomomorphismByImages(G, Q, GeneratorsOfGroup(G), GeneratorsOfGroup(quo_rec.Image));
   ImagesOfGens := List(GeneratorsOfGroup(G), g -> Image(rep, Image(f, g)));
  return RankOfHomologyWithLocalCoeffs(G, ImagesOfGens);
end;

ComputeDataOnModularRepresentation :=function(manifolds, p, PrintFun)
   local ident, Q, M, quo_rec, reps, i, dims, ranks, adj;
   ident := [(p^3 - p)/2, 1];
   Q := SimpleGroup(ident);
   reps := ModularRepresentations(Q, p);
   dims := List(reps, r -> DimensionsMat(r[1])[1]);
   adj := Position(dims, p-2);
   PrintFun("Dims  ", dims, "\n\n");
   for M in manifolds do
     i := 1;
     for quo_rec in M.Quotients do
	if quo_rec.Type = "simple" and quo_rec.Ident = ident then
	    PrintFun(M.Name, ":", i, "  Qrk ", quo_rec.Rank, "  ");
            ranks := List(reps, rep -> 
                   ComputeTwistedHomology(M, Q, quo_rec, rep));
            # adjust because H^1(PSL(2,p), V_(p-2)) = 1
            ranks[adj] := ranks[adj] - 1;
	    PrintFun(ranks);
            PrintFun("\n");
            i := i + 1;
         fi;
     od;
   od;
end;


   