import math, sys, string

# Basic Statistics functions

def mean(data):
    sum = 0.0
    for d in data: sum = sum + d
    return sum/len(data)

def median(data):
    y = data[:]
    y.sort()
    n = len(y)/2
    if len(y) % 2 == 0:
        return 0.5*(y[n-1] + y[n])
    else:
        return y[n]

# note: below only works for x, y a sequence of 0s and 1s

def correlation(x, y):
    assert(len(x) == len(y))
    if x.count(1) == 0 or x.count(0) == 0:
        return "inf"
    if y.count(1) == 0 or y.count(0) == 0:
        return "inf"
    xmean, ymean = mean(x), mean(y)
    xymean = mean( map(lambda i,x=x,y=y: x[i]*y[i], range(len(x))))
    ans = (xymean - xmean*ymean)/math.sqrt(
       (xmean - xmean**2)*(ymean - ymean**2))
    return ans

def make_correlations(raw_data):
    cor = []
    for dat1 in raw_data:
        cor_line = []
        for dat2 in raw_data:
            cor_line.append(correlation(dat1, dat2))
        cor.append(cor_line)
    return cor
    

# 3-manifold specific part

def load_manifold_data(filename):
    return eval(open(filename).read())
    
def has_cover_correlations(data):
    raw_data = map(lambda x: x[2], data)
    def make_abs(x):
        return map(lambda y: abs(y), x)
    return make_correlations(map(make_abs, raw_data))

def has_cover_for_mathematica(data):
    raw_data = map(lambda x: x[2], data)
    def make_abs(x):
        return map(lambda y: abs(y), x)
    final_data = map(make_abs, raw_data)
    out_str = string.replace(repr(final_data), "[", "{")
    return string.replace(out_str, "]", "}")

def hit_correlations(data):
    raw_data = map(lambda x: x[2], data)
    def hit(x):
        def h(y):
            if y > 0:
                return 1
            else:
                return 0
        return map(h, x)
    return make_correlations(map(hit, raw_data))

def hit_correlations_for_mathematica(data):
    raw_data = map(lambda x: x[2], data)
    def hit(x):
        def h(y):
            if y > 0:
                return 1
            else:
                return 0
        return map(h, x)
    final_data = map(hit, raw_data)
    out_str = string.replace(repr(final_data), "[", "{")
    return string.replace(out_str, "]", "}")

# In the following function, we focus only those data for which
# the manifold had both kinds of covers.

def success_failure_correlations(data):
    def f(n):
        if n > 0: return 1
        else: return 0

    raw_data = map(lambda x: x[2], data)
    cor = []
    for dat1 in raw_data:
        cor_line = []
        for dat2 in raw_data:
            x, y = [], []
            for i in range(len(dat1)):
                if dat1[i] * dat2[i] != 0:
                    x.append(f(dat1[i]))
                    y.append(f(dat2[i]))
            cor_line.append(correlation(x,y))
        cor.append(cor_line)
    return cor

# below functions are for displaying correlation data

#pads tail

def make_len(str, l):
    return str + " "*max(l - len(str), 0)
        
def format_percentage(percent, spacing):
    str = "%.1f" % percent
    return " "*max(5 - len(str), 0) + str + spacing

def format_cor(cor, spacing):
    if cor == "inf":
        str = cor
    else:
        str = "%.2f" % cor
    return " "*max(5 - len(str), 0) + str + spacing


def display_correlations(data, func, columns = None):
    # compute stats
    extra_data = []
    for line in data:
        hits = 0.0
        covers = 0.0
        n = 0
        for e in line[2]:
            if e == 1:
                hits = hits + 1
                covers = covers + 1
            elif e == -1:
                covers = covers + 1
            n = n + 1

        if covers:
            extra_data.append((100*hits/n, 100*covers/n, 100*hits/covers))
        else:
            extra_data.append((0,0,0))

    cor = func(data)
    if columns == None:
        columns = range(len(cor[0]))

    # print header

    o = sys.stdout.write

    o("         Hit   HavCov SucRat    ")
    for j in columns:
        o(make_len(data[j][1], 6)+ "  ")

    o("\n")

    # print each line
    for i in range(len(data)):
        o(make_len(data[i][1], 6) + "  ")
        for per in extra_data[i]:
            o(format_percentage(per, "  "))
        o("  ")
        for j in columns:
            o(format_cor(cor[i][j], "   "))
        o("\n")

def print_full_report(data, ranges=None):
    o = sys.stdout.write
    o("    CORRELATION BETWEEN MANIFOLDS HAVING A COVER WITH GIVEN GROUP\n\n")
    if ranges == None:
        display_correlations(data, has_cover_correlations)
    else:
        for cols in ranges:
            display_correlations(data, has_cover_correlations, cols)
            o("\n")

    o("    CORRELATION BETWEEN MANIFOLDS HAVING A HIT WITH GIVEN GROUP\n\n")
    if ranges == None:
        display_correlations(data, hit_correlations)
    else:
        for cols in ranges:
            display_correlations(data, hit_correlations, cols)
            o("\n")

    o("\n")
    o("    CORRELATION BETWEEN MANIFOLDS HAVING A HIT WITH GIVEN GROUP\n")
    o("      Ingoring manifolds which do not have a cover with both of\n")
    o("      the groups being compared. (Data is thus small, prob. too\n")
    o("      small).\n\n")
    if ranges == None:
        display_correlations(data, success_failure_correlations)
    else:
        for cols in ranges:
            display_correlations(data, success_failure_correlations, cols)
            o("\n")
            

def number_found_through_index(data):
    out_data = []
    L = len(data[0][2])
    finished = [0,]*L
    for D in data:
        order = D[0][0]
        for i in range(L):
            if D[2][i] > 0:
                finished[i] = 1
        total = finished.count(1)
        out_data.append([order, total])
    return out_data
        
def number_found_through_index_psl_only(data):
    out_data = []
    L = len(data[0][2])
    finished = [0,]*L
    for D in data:
        if D[1][0] != "L" and D[1] != "A5":
            continue
        order = D[0][0]
        for i in range(L):
            if D[2][i] > 0:
                finished[i] = 1
        total = finished.count(1)
        out_data.append([order, total])
    return out_data
        
        
def make_simple_group_figure():
    import Gnuplot, math
    d = load_manifold_data("simple1000.stat")[:-1]
    n1 = number_found_through_index(d)
    n1 = map(lambda x,f=math.log10:[f(x[0]), x[1]/10.0], n1)
    g = Gnuplot.Gnuplot()
    g('set term postscript eps monochrome "Times-Roman" 18')
    g('set output "simple_groups.eps"')
    g("set yrange [0:100]")
    g("set data style linespoints")
    g.xlabel("x = Log(Order of group)")
    g.ylabel("% with a pos. betti. num. simple cover of degree < x")
    g("set pointsize 2")
    g.plot(n1)

def make_homology_figure():
    import Gnuplot, math
    d = eval(open("simple_dat").read())
    n1 = []
    for l in d:
        #n1.append([math.log10(l[0][0]/l[1]), math.log10(mean(l[2]))])
        n1.append([l[0][0]/l[1], mean(l[2])])
    return n1
    #g = Gnuplot.Gnuplot()
    #g.plot(n1)
    #return g
    #g('set term postscript eps monochrome "Times-Roman" 18')
    #g('set output "homology.eps"')
    #g.xlabel("Log(degree of cover)")
    #g.ylabel("Log(mean rank of H_1)")
    #g.plot(n1)

def hit_rates(data):
    hit_r = []
    for line in data:
        hits = 0.0
        n = 0
        for e in line[2]:
            if e == 1:
                hits = hits + 1
            n = n + 1
        
        hit_r.append(hits/n)
    return hit_r

def hit_rate_model(data):
    curr = 0
    real = 0
    done = []
    err = []
    N = number_found_through_index(d)
    H = hit_rates(data)
    for i in range(len(H)):
        curr = curr + (1000 - curr)*H[i]
        done.append([N[i][0], curr])

    V = map(lambda x:x[1], N)
    for i in range(1, len(H)):
        #err.append( (V[i - 1] + (1000 - V[i - 1])*H[i] - V[i])/V[i] )
        err.append( (V[i - 1] + (1000 - V[i - 1])*H[i] - V[i])/(1 - V[i-1]) )

    return [done, N, err]
        
def make_simple_group_figure2():
    import Gnuplot, math
    d = load_manifold_data("simple1000.stat")[:-1]
    n1 = number_found_through_index(d)
    n1 = map(lambda x,f=math.log10:[f(x[0]), x[1]/10.0], n1)
    n2 = number_found_through_index_psl_only(d)
    n2 = map(lambda x,f=math.log10:[f(x[0]), x[1]/10.0], n2)
    n3,n,e = hit_rate_model(d)
    n3 = map(lambda x,f=math.log10:[f(x[0]), x[1]/10.0], n3)
    g = Gnuplot.Gnuplot()
    g('set term postscript eps monochrome "Times-Roman" 18')
    g('set output "simple_groups2.eps"')
    g("set yrange [0:100]")
    g("set data style linespoints")
    g.xlabel("x = Log(Order of group)")
    g.ylabel("% with a pos. betti. num. simple cover of degree < x")
    g("set pointsize 2")
    g.plot(n1, n2, n3)

def measure_correlation_drift(data, func):
    c = func(data)
    s = 0
    for cc in c:
        for ccc in cc:
            s = s + ccc
    n = len(data)
    assert n == len(c)
    return [s, (s - n) / (n**2 - 1)]
    
if __name__ == "__main__":
    None
    #d = load_manifold_data("simple1000.stat")[:-1]
    #cut off L2(41)

    #d = load_manifold_data("fig8.stat")
    #display_correlations(d, has_cover_correlations, range(0,10))
    #print "\n"
    #display_correlations(d, has_cover_correlations, range(10,19))

    #d = load_manifold_data("fig8.stat")
    #display_correlations(d,has_cover_correlations)
    #d = load_manifold_data("whitehead8.stat")
    #display_correlations(d, has_cover_correlations, range(0,12))
    #print "\n"
    #display_correlations(d, has_cover_correlations, range(12,24))
                         
    #save_correlations(d, has_cover_correlations, "test.tex")
    #print
    #print d
    #display_correlations(d, hit_correlations)
    #print
    #display_correlations(d, success_failure_correlations)
    #print hit_correlations_for_mathematica(d)
        
           
    
